@@ -4,71 +4,88 @@ import fetch from 'node-fetch'
 import * as URI from 'urijs'
 import * as url from 'url'
 import { promisify } from 'util'
-import * as LSP from 'vscode-languageserver'
+import * as LSP from 'vscode-languageserver/node'
+import { TextDocument } from 'vscode-languageserver-textdocument'
 import * as Parser from 'web-tree-sitter'
 
-import * as config from './config'
 import { flattenArray, flattenObjectValues } from './util/flatten'
 import { getFilePaths } from './util/fs'
-import { getShebang, isBashShebang } from './util/shebang'
+import { analyzeShebang } from './util/shebang'
+import * as sourcing from './util/sourcing'
 import * as TreeSitterUtil from './util/tree-sitter'
 
 const readFileAsync = promisify(fs.readFile)
 
-type Kinds = { [type: string]: LSP.SymbolKind }
+const TREE_SITTER_TYPE_TO_LSP_KIND: { [type: string]: LSP.SymbolKind | undefined } = {
+  // These keys are using underscores as that's the naming convention in tree-sitter.
+  environment_variable_assignment: LSP.SymbolKind.Variable,
+  function_definition: LSP.SymbolKind.Function,
+  variable_assignment: LSP.SymbolKind.Variable,
+}
 
 type Declarations = { [word: string]: LSP.SymbolInformation[] }
-type FileDeclarations = { [uri: string]: Declarations }
 
-type Trees = { [uri: string]: Parser.Tree }
-type Texts = { [uri: string]: string }
+type AnalyzedDocument = {
+  document: TextDocument
+  tree: Parser.Tree
+  declarations: Declarations
+  sourcedUris: Set<string>
+}
 
 /**
  * The Analyzer uses the Abstract Syntax Trees (ASTs) that are provided by
  * tree-sitter to find definitions, reference, etc.
  */
 export default class Analyzer {
-  private parser: Parser
   private console: LSP.RemoteConsole
-
-  private uriToTextDocument: { [uri: string]: LSP.TextDocument } = {}
-
-  private uriToTreeSitterTrees: Trees = {}
-
-  // We need this to find the word at a given point etc.
-  private uriToFileContent: Texts = {}
-
-  private uriToDeclarations: FileDeclarations = {}
-
-  private treeSitterTypeToLSPKind: Kinds = {
-    // These keys are using underscores as that's the naming convention in tree-sitter.
-    environment_variable_assignment: LSP.SymbolKind.Variable,
-    function_definition: LSP.SymbolKind.Function,
-    variable_assignment: LSP.SymbolKind.Variable,
-  }
+  private includeAllWorkspaceSymbols: boolean
+  private parser: Parser
+  private uriToAnalyzedDocument: Record<string, AnalyzedDocument | undefined> = {}
+  private workspaceFolder: string | null
 
   public constructor({
     console,
+    includeAllWorkspaceSymbols = false,
     parser,
+    workspaceFolder,
   }: {
     console: LSP.RemoteConsole
+    includeAllWorkspaceSymbols?: boolean
     parser: Parser
+    workspaceFolder: string | null
   }) {
-    this.parser = parser
     this.console = console
+    this.includeAllWorkspaceSymbols = includeAllWorkspaceSymbols
+    this.parser = parser
+    this.workspaceFolder = workspaceFolder
   }
 
   /**
-   * Initiates a background analysis of the files in the given rootPath to
+   * Initiates a background analysis of the files in the workspaceFolder to
    * enable features across files.
+   *
+   * NOTE that when the source aware feature is enabled files are also parsed
+   * when they are found.
    */
   public async initiateBackgroundAnalysis({
-    rootPath,
+    backgroundAnalysisMaxFiles,
+    globPattern,
   }: {
-    rootPath: string
+    backgroundAnalysisMaxFiles: number
+    globPattern: string
   }): Promise<{ filesParsed: number }> {
-    const globPattern = config.getGlobPattern()
-    const backgroundAnalysisMaxFiles = config.getBackgroundAnalysisMaxFiles()
+    const rootPath = this.workspaceFolder
+    if (!rootPath) {
+      return { filesParsed: 0 }
+    }
+
+    if (backgroundAnalysisMaxFiles <= 0) {
+      this.console.log(
+        `BackgroundAnalysis: skipping as backgroundAnalysisMaxFiles was 0...`,
+      )
+      return { filesParsed: 0 }
+    }
+
     this.console.log(
       `BackgroundAnalysis: resolving glob "${globPattern}" inside "${rootPath}"...`,
     )
@@ -102,15 +119,18 @@ export default class Analyzer {
 
       try {
         const fileContent = await readFileAsync(filePath, 'utf8')
-        const shebang = getShebang(fileContent)
-        if (shebang && !isBashShebang(shebang)) {
+        const { shebang, shellDialect } = analyzeShebang(fileContent)
+        if (shebang && !shellDialect) {
           this.console.log(
             `BackgroundAnalysis: Skipping file ${uri} with shebang "${shebang}"`,
           )
           continue
         }
 
-        this.analyze(uri, LSP.TextDocument.create(uri, 'shell', 1, fileContent))
+        this.analyze({
+          document: TextDocument.create(uri, 'shell', 1, fileContent),
+          uri,
+        })
       } catch (error) {
         const errorMessage = error instanceof Error ? error.message : error
         this.console.warn(
@@ -128,13 +148,41 @@ export default class Analyzer {
   /**
    * Find all the locations where something has been defined.
    */
-  public findDefinition({ word }: { word: string }): LSP.Location[] {
-    const symbols: LSP.SymbolInformation[] = []
-    Object.keys(this.uriToDeclarations).forEach((uri) => {
-      const declarationNames = this.uriToDeclarations[uri][word] || []
-      declarationNames.forEach((d) => symbols.push(d))
-    })
-    return symbols.map((s) => s.location)
+  public findDefinition({
+    position,
+    uri,
+    word,
+  }: {
+    position?: { line: number; character: number }
+    uri: string
+    word: string
+  }): LSP.Location[] {
+    const tree = this.uriToAnalyzedDocument[uri]?.tree
+
+    if (position && tree) {
+      // NOTE: when a word is a file path to a sourced file, we return a location to it.
+      const sourcedLocation = sourcing.getSourcedLocation({
+        position,
+        rootPath: this.workspaceFolder,
+        tree,
+        uri,
+        word,
+      })
+      if (sourcedLocation) {
+        return [sourcedLocation]
+      }
+    }
+
+    return this.getReachableUris({ uri })
+      .reduce((symbols, uri) => {
+        const analyzedDocument = this.uriToAnalyzedDocument[uri]
+        if (analyzedDocument) {
+          const declarationNames = analyzedDocument.declarations[word] || []
+          declarationNames.forEach((d) => symbols.push(d))
+        }
+        return symbols
+      }, [] as LSP.SymbolInformation[])
+      .map((symbol) => symbol.location)
   }
 
   /**
@@ -154,13 +202,17 @@ export default class Analyzer {
     params: LSP.TextDocumentPositionParams
     endpoint: string
   }): Promise<{ helpHTML?: string }> {
-    const leafNode = this.uriToTreeSitterTrees[
-      params.textDocument.uri
-    ].rootNode.descendantForPosition({
+    const analyzedDocument = this.uriToAnalyzedDocument[params.textDocument.uri]
+
+    const leafNode = analyzedDocument?.tree.rootNode.descendantForPosition({
       row: params.position.line,
       column: params.position.character,
     })
 
+    if (!leafNode || !analyzedDocument) {
+      return {}
+    }
+
     // explainshell needs the whole command, not just the "word" (tree-sitter
     // parlance) that the user hovered over. A relatively successful heuristic
     // is to simply go up one level in the AST. If you go up too far, you'll
@@ -172,10 +224,9 @@ export default class Analyzer {
       return {}
     }
 
-    const cmd = this.uriToFileContent[params.textDocument.uri].slice(
-      interestingNode.startIndex,
-      interestingNode.endIndex,
-    )
+    const cmd = analyzedDocument.document
+      .getText()
+      .slice(interestingNode.startIndex, interestingNode.endIndex)
     type ExplainshellResponse = {
       matches?: Array<{ helpHTML: string; start: number; end: number }>
     }
@@ -191,8 +242,7 @@ export default class Analyzer {
       return {}
     } else {
       const offsetOfMousePointerInCommand =
-        this.uriToTextDocument[params.textDocument.uri].offsetAt(params.position) -
-        interestingNode.startIndex
+        analyzedDocument.document.offsetAt(params.position) - interestingNode.startIndex
 
       const match = explainshellResponse.matches.find(
         (helpItem) =>
@@ -208,7 +258,7 @@ export default class Analyzer {
    * Find all the locations where something named name has been defined.
    */
   public findReferences(name: string): LSP.Location[] {
-    const uris = Object.keys(this.uriToTreeSitterTrees)
+    const uris = Object.keys(this.uriToAnalyzedDocument)
     return flattenArray(uris.map((uri) => this.findOccurrences(uri, name)))
   }
 
@@ -217,8 +267,13 @@ export default class Analyzer {
    * It's currently not scope-aware.
    */
   public findOccurrences(uri: string, query: string): LSP.Location[] {
-    const tree = this.uriToTreeSitterTrees[uri]
-    const contents = this.uriToFileContent[uri]
+    const analyzedDocument = this.uriToAnalyzedDocument[uri]
+    if (!analyzedDocument) {
+      return []
+    }
+
+    const { tree } = analyzedDocument
+    const contents = analyzedDocument.document.getText()
 
     const locations: LSP.Location[] = []
 
@@ -250,35 +305,36 @@ export default class Analyzer {
    * Find all symbol definitions in the given file.
    */
   public findSymbolsForFile({ uri }: { uri: string }): LSP.SymbolInformation[] {
-    const declarationsInFile = this.uriToDeclarations[uri] || {}
+    const declarationsInFile = this.uriToAnalyzedDocument[uri]?.declarations || {}
     return flattenObjectValues(declarationsInFile)
   }
 
   /**
    * Find symbol completions for the given word.
-   *
-   * TODO: if a file is not included we probably shouldn't include it declarations from it.
    */
   public findSymbolsMatchingWord({
     exactMatch,
+    uri,
     word,
   }: {
     exactMatch: boolean
+    uri: string
     word: string
   }): LSP.SymbolInformation[] {
-    const symbols: LSP.SymbolInformation[] = []
+    return this.getReachableUris({ uri }).reduce((symbols, uri) => {
+      const analyzedDocument = this.uriToAnalyzedDocument[uri]
 
-    Object.keys(this.uriToDeclarations).forEach((uri) => {
-      const declarationsInFile = this.uriToDeclarations[uri] || {}
-      Object.keys(declarationsInFile).map((name) => {
-        const match = exactMatch ? name === word : name.startsWith(word)
-        if (match) {
-          declarationsInFile[name].forEach((symbol) => symbols.push(symbol))
-        }
-      })
-    })
-
-    return symbols
+      if (analyzedDocument) {
+        const { declarations } = analyzedDocument
+        Object.keys(declarations).map((name) => {
+          const match = exactMatch ? name === word : name.startsWith(word)
+          if (match) {
+            declarations[name].forEach((symbol) => symbols.push(symbol))
+          }
+        })
+      }
+      return symbols
+    }, [] as LSP.SymbolInformation[])
   }
 
   /**
@@ -288,18 +344,22 @@ export default class Analyzer {
    * Returns all, if any, syntax errors that occurred while parsing the file.
    *
    */
-  public analyze(uri: string, document: LSP.TextDocument): LSP.Diagnostic[] {
+  public analyze({
+    document,
+    uri, // NOTE: we don't use document.uri to make testing easier
+  }: {
+    document: TextDocument
+    uri: string
+  }): LSP.Diagnostic[] {
     const contents = document.getText()
 
     const tree = this.parser.parse(contents)
 
-    this.uriToTextDocument[uri] = document
-    this.uriToTreeSitterTrees[uri] = tree
-    this.uriToDeclarations[uri] = {}
-    this.uriToFileContent[uri] = contents
-
     const problems: LSP.Diagnostic[] = []
 
+    const declarations: Declarations = {}
+
+    // TODO: move this somewhere
     TreeSitterUtil.forEach(tree.rootNode, (n: Parser.SyntaxNode) => {
       if (n.type === 'ERROR') {
         problems.push(
@@ -318,7 +378,7 @@ export default class Analyzer {
         }
 
         const word = contents.slice(named.startIndex, named.endIndex)
-        const namedDeclarations = this.uriToDeclarations[uri][word] || []
+        const namedDeclarations = declarations[word] || []
 
         const parent = TreeSitterUtil.findParent(
           n,
@@ -332,19 +392,39 @@ export default class Analyzer {
               )
             : '' // TODO: unsure what we should do here?
 
+        const kind = TREE_SITTER_TYPE_TO_LSP_KIND[n.type]
+
+        if (!kind) {
+          this.console.warn(
+            `Unmapped tree sitter type: ${n.type}, defaulting to variable`,
+          )
+        }
+
         namedDeclarations.push(
           LSP.SymbolInformation.create(
             word,
-            this.treeSitterTypeToLSPKind[n.type],
+            kind || LSP.SymbolKind.Variable,
             TreeSitterUtil.range(n),
             uri,
             parentName,
           ),
         )
-        this.uriToDeclarations[uri][word] = namedDeclarations
+        declarations[word] = namedDeclarations
       }
     })
 
+    this.uriToAnalyzedDocument[uri] = {
+      tree,
+      document,
+      declarations,
+      sourcedUris: sourcing.getSourcedUris({
+        fileContent: contents,
+        fileUri: uri,
+        rootPath: this.workspaceFolder,
+        tree,
+      }),
+    }
+
     function findMissingNodes(node: Parser.SyntaxNode) {
       if (node.isMissing()) {
         problems.push(
@@ -364,6 +444,29 @@ export default class Analyzer {
     return problems
   }
 
+  public findAllSourcedUris({ uri }: { uri: string }): Set<string> {
+    const allSourcedUris = new Set<string>([])
+
+    const addSourcedFilesFromUri = (fromUri: string) => {
+      const sourcedUris = this.uriToAnalyzedDocument[fromUri]?.sourcedUris
+
+      if (!sourcedUris) {
+        return
+      }
+
+      sourcedUris.forEach((sourcedUri) => {
+        if (!allSourcedUris.has(sourcedUri)) {
+          allSourcedUris.add(sourcedUri)
+          addSourcedFilesFromUri(sourcedUri)
+        }
+      })
+    }
+
+    addSourcedFilesFromUri(uri)
+
+    return allSourcedUris
+  }
+
   /**
    * Find the node at the given point.
    */
@@ -372,14 +475,14 @@ export default class Analyzer {
     line: number,
     column: number,
   ): Parser.SyntaxNode | null {
-    const document = this.uriToTreeSitterTrees[uri]
+    const tree = this.uriToAnalyzedDocument[uri]?.tree
 
-    if (!document?.rootNode) {
+    if (!tree?.rootNode) {
       // Check for lacking rootNode (due to failed parse?)
       return null
     }
 
-    return document.rootNode.descendantForPosition({ row: line, column })
+    return tree.rootNode.descendantForPosition({ row: line, column })
   }
 
   /**
@@ -422,7 +525,10 @@ export default class Analyzer {
    * Find a block of comments above a line position
    */
   public commentsAbove(uri: string, line: number): string | null {
-    const doc = this.uriToTextDocument[uri]
+    const doc = this.uriToAnalyzedDocument[uri]?.document
+    if (!doc) {
+      return null
+    }
 
     const commentBlock = []
 
@@ -469,19 +575,54 @@ export default class Analyzer {
     return null
   }
 
-  public getAllVariableSymbols(): LSP.SymbolInformation[] {
-    return this.getAllSymbols().filter(
+  public getAllVariableSymbols({ uri }: { uri: string }): LSP.SymbolInformation[] {
+    return this.getAllSymbols({ uri }).filter(
       (symbol) => symbol.kind === LSP.SymbolKind.Variable,
     )
   }
 
-  private getAllSymbols(): LSP.SymbolInformation[] {
-    // NOTE: this could be cached, it takes < 1 ms to generate for a project with 250 bash files...
-    const symbols: LSP.SymbolInformation[] = []
+  public setIncludeAllWorkspaceSymbols(includeAllWorkspaceSymbols: boolean): void {
+    this.includeAllWorkspaceSymbols = includeAllWorkspaceSymbols
+  }
 
-    Object.keys(this.uriToDeclarations).forEach((uri) => {
-      Object.keys(this.uriToDeclarations[uri]).forEach((name) => {
-        const declarationNames = this.uriToDeclarations[uri][name] || []
+  private getReachableUris({ uri: fromUri }: { uri?: string } = {}): string[] {
+    if (!fromUri || this.includeAllWorkspaceSymbols) {
+      return Object.keys(this.uriToAnalyzedDocument)
+    }
+
+    const uris = [fromUri, ...Array.from(this.findAllSourcedUris({ uri: fromUri }))]
+
+    return uris.filter((uri) => {
+      if (!this.uriToAnalyzedDocument[uri]) {
+        // Either the background analysis didn't run or the file is outside
+        // the workspace. Let us try to analyze the file.
+        try {
+          const fileContent = fs.readFileSync(new URL(uri), 'utf8')
+          this.analyze({
+            document: TextDocument.create(uri, 'shell', 1, fileContent),
+            uri,
+          })
+        } catch (err) {
+          this.console.log(`Error while analyzing sourced file ${uri}: ${err}`)
+          return false
+        }
+      }
+
+      return true
+    })
+  }
+
+  private getAllSymbols({ uri }: { uri?: string } = {}): LSP.SymbolInformation[] {
+    const reachableUris = this.getReachableUris({ uri })
+
+    const symbols: LSP.SymbolInformation[] = []
+    reachableUris.forEach((uri) => {
+      const analyzedDocument = this.uriToAnalyzedDocument[uri]
+      if (!analyzedDocument) {
+        return
+      }
+
+      Object.values(analyzedDocument.declarations).forEach((declarationNames) => {
         declarationNames.forEach((d) => symbols.push(d))
       })
     })
